/***************************************************************************
 *   Copyright (C) 2007 by David Lopez Vinacua                             *
 *   david@bajavista.com                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

/* * * * * * * * * * * * * * * * * * * * *
 *  tdview.c                             *
 *  Code and Data for tinydns log viewer *
 * * * * * * * * * * * * * * * * * * * * *
 *
 * Created:
 *  260907 David Lopez
 * Modified:
 *
 */

#include <stdio.h>
#include <time.h>
#include <string.h>

#define LogPath "/service/tinydns/log/main"
#define Current "current"

typedef unsigned char Byte;

short daltool_GetHexByte(char *src)
/* Gets a two-digit hex byte from src. Returns -1 if fails */
{
  Byte          v,w;

  if (!src || !(*src))
     return -1;
  if ( (*src>='0') && (*src<='9') )
     v=(*src)-'0';
  else if ( (*src>='a') && (*src<='f') )
     v=(*src)-'a'+10;
  else if ( (*src>='A') && (*src<='F') )
     v=(*src)-'A'+10;
  else
     return -1;
  src++;
  if (!*src)
     return v;
  if ( (*src>='0') && (*src<='9') )
     w=(*src)-'0';
  else if ( (*src>='a') && (*src<='f') )
     w=(*src)-'a'+10;
  else if ( (*src>='A') && (*src<='F') )
     w=(*src)-'A'+10;
  else
     return -1;
  return 16*v+w;
}

short GetHexu32(char *src, unsigned int *res)
/* Gets an unsigned 32bit Hex BigEndian value, returns 0 on success */
{
  unsigned int    v;
  short           r,i;
  
  if (!src)
     return -1;
  v=0;
  i=0;
  while (i<4)
    {
    r=daltool_GetHexByte(src);
    if (r>=0)
       {
       v=256*v+r;
       src+=2;
       }
    else
       return -2;
    i++;
    }
  *res=v;
  return 0;
}

short GetHexu16(char *src, unsigned short *res)
/* Gets an unsigned 16bit Hex BigEndian value, returns 0 on success */
{
  unsigned short  v;
  short           r,i;
  
  if (!src)
     return -1;
  v=0;
  i=0;
  while (i<2)
    {
    r=daltool_GetHexByte(src);
    if (r>=0)
       {
       v=256*v+r;
       src+=2;
       }
    else
       return -2;
    i++;
    }
  *res=v;
  return 0;
}

char *Parse_Type(unsigned short Type)
/* Returns query type */
{
  static char     s[8];
  
  switch (Type)
    {
    case 0x0001: return "A";
    case 0x0002: return "NS";
    case 0x0005: return "CNAME";
    case 0x0006: return "SOA";
    case 0x000c: return "PTR";
    case 0x000f: return "MX";
    case 0x0010: return "TXT";
    case 0x001b: return "GPOS";
    case 0x001c: return "AAAA";
    case 0x0026: return "A6";
    case 0x0063: return "SPF";
    case 0x00fb: return "IXFR";
    case 0x00fc: return "AXFR";
    case 0x00ff: return "*";
    default: 
      sprintf(s,"[0x%04x]",Type);
      return &s[0];
    }
}

short Parse(FILE *f, FILE *o)
/* Parses given file to standard output */
{
  char            s[256];
  unsigned short  cp,pp;
  unsigned int    i1,i2;
  unsigned short  w1,w2;
  time_t          t;
  struct tm      *tm;
  
  if (f)
     {
     while ( (fgets(&s[0],sizeof(s),f)) )
       {
       cp=0;
       // Line starts with the time stamp
       if (s[0]=='@')
          {
          cp++;
          pp=cp;
          while (s[cp] && (s[cp]!=' '))
            cp++;
          // Timestamp is 24 hex digits => 3x32bit values
          // It seems to use 32bit nanosecond value at the end (9 digits)
          // The rest may be a 64bit time_t, but only lower 32bit seem valid
          // So we ignore first 32bit value, by now
          if ( (cp-pp==24) && s[cp] )
             {
             if (GetHexu32(&s[pp+8],&i1) || GetHexu32(&s[pp+16],&i2))
                fprintf(o,"**Invalid line (wrong timestamp)**\n");
             else
                {
                t=i1;
                tm=localtime(&t);
                fprintf(o,"%04u%02u%02u-%02u:%02u:%02u.%06u ",
                 1900+tm->tm_year,1+tm->tm_mon,tm->tm_mday,tm->tm_hour,tm->tm_min,tm->tm_sec,i2/1000);
                // Now comes the logged message
                cp++;
                pp=cp;
                cp=strlen(&s[0]);
                if (cp && (s[cp-1]=='\n'))
                   s[cp-1]='\0';
                if (s[pp]=='s') // "starting tinydns"
                   fprintf(o,"%s\n",&s[pp]);
                else
                   {
                   if (!GetHexu32(&s[pp],&i1) && !GetHexu16(&s[pp+9],&w1) && !GetHexu16(&s[pp+14],&w2))
                      { // Got IP:port:id
                      fprintf(o,"%u.%u.%u.%u:%u ",
                       daltool_GetHexByte(&s[pp]),daltool_GetHexByte(&s[pp+2]),daltool_GetHexByte(&s[pp+4]),daltool_GetHexByte(&s[pp+6]),w1);
                      pp+=19;
                      GetHexu16(&s[pp+2],&w1);
                      cp=pp+7;
                      switch (s[pp])
                        {
                        case '+':
                          fprintf(o,"OK %s (0x%04x) %s\n",Parse_Type(w1),w2,&s[cp]); break;
                        case '-':
                          fprintf(o,"IGNORED (NoAuth) %s (0x%04x) %s\n",Parse_Type(w1),w2,&s[cp]); break;
                        case 'I':
                          fprintf(o,"IGNORED (NotImplQry) %s (0x%04x) %s\n",Parse_Type(w1),w2,&s[cp]); break;
                        case 'C':
                          fprintf(o,"IGNORED (NotImplClass) %s (0x%04x) %s\n",Parse_Type(w1),w2,&s[cp]); break;
                        case '/':
                          fprintf(o,"DROPPED\n"); break;
                        default:
                          fprintf(o,"Unknown tinydns type [%c]\n",s[pp]);
                        }
                      }
                   else
                      fprintf(o,"**Invalid data in line**\n");
                   }
                }
             }
          else
             fprintf(o,"**Invalid line (bad timestamp)**\n");
          }
       else
          fprintf(o,"**Invalid line**\n");
       }
     fclose(f);
     }
  return 0;
}

int main(int argc, char *argv[])
{
  FILE           *f,*o;
  char            fn[192];
  
  o=stdout;
  if (argc>1)
     {
     if (strcmp(argv[1],"-")==0)
        f=stdin;
     else
        {
        fprintf(stdout,"Usage:\n %s [-]\n With no params shows default %s/%s,\n with a dash takes input from standard input\n",
         argv[0],LogPath,Current);
        return 0;
        }
     }
  else
     {
     sprintf(fn,"%s/%s",LogPath,Current);
     f=fopen(&fn[0],"r");
     }
  Parse(f,o);
  return 0;
}
